@@ -5,9 +5,10 @@ from typing import (
     Collection,
     Dict,
     Generic,
-    List,
+    Mapping,
     NamedTuple,
     Optional,
+    Tuple,
     TYPE_CHECKING,
     Type,
     TypeVar,
@@ -44,7 +45,6 @@ from ..language import (
 )
 from ..pyutils import (
     AwaitableOrValue,
-    FrozenList,
     Path,
     cached_property,
     did_you_mean,
@@ -55,6 +55,12 @@ from ..pyutils import (
     Undefined,
 )
 from ..utilities.value_from_ast_untyped import value_from_ast_untyped
+from .assert_name import assert_name, assert_enum_value_name
+
+try:
+    from typing import TypedDict
+except ImportError:  # Python < 3.8
+    from typing_extensions import TypedDict
 
 if TYPE_CHECKING:
     from .schema import GraphQLSchema  # noqa: F401
@@ -98,39 +104,55 @@ __all__ = [
     "assert_named_type",
     "get_nullable_type",
     "get_named_type",
+    "resolve_thunk",
     "GraphQLAbstractType",
     "GraphQLArgument",
+    "GraphQLArgumentKwargs",
     "GraphQLArgumentMap",
     "GraphQLCompositeType",
     "GraphQLEnumType",
+    "GraphQLEnumTypeKwargs",
     "GraphQLEnumValue",
+    "GraphQLEnumValueKwargs",
     "GraphQLEnumValueMap",
     "GraphQLField",
+    "GraphQLFieldKwargs",
     "GraphQLFieldMap",
     "GraphQLFieldResolver",
     "GraphQLInputField",
+    "GraphQLInputFieldKwargs",
     "GraphQLInputFieldMap",
     "GraphQLInputObjectType",
+    "GraphQLInputObjectTypeKwargs",
     "GraphQLInputType",
     "GraphQLInterfaceType",
+    "GraphQLInterfaceTypeKwargs",
     "GraphQLIsTypeOfFn",
     "GraphQLLeafType",
     "GraphQLList",
     "GraphQLNamedType",
+    "GraphQLNamedTypeKwargs",
+    "GraphQLNamedInputType",
+    "GraphQLNamedOutputType",
     "GraphQLNullableType",
     "GraphQLNonNull",
     "GraphQLResolveInfo",
     "GraphQLScalarType",
+    "GraphQLScalarTypeKwargs",
     "GraphQLScalarSerializer",
     "GraphQLScalarValueParser",
     "GraphQLScalarLiteralParser",
     "GraphQLObjectType",
+    "GraphQLObjectTypeKwargs",
     "GraphQLOutputType",
     "GraphQLType",
     "GraphQLTypeResolver",
     "GraphQLUnionType",
+    "GraphQLUnionTypeKwargs",
     "GraphQLWrappingType",
     "Thunk",
+    "ThunkCollection",
+    "ThunkMapping",
 ]
 
 
@@ -186,7 +208,14 @@ def assert_wrapping_type(type_: Any) -> GraphQLWrappingType:
     return cast(GraphQLWrappingType, type_)
 
 
-# These named types do not include modifiers like List or NonNull.
+class GraphQLNamedTypeKwargs(TypedDict, total=False):
+    name: str
+    description: Optional[str]
+    extensions: Dict[str, Any]
+    # unfortunately, we cannot make the following more specific, because they are
+    # used by subclasses with different node types and typed dicts cannot be refined
+    ast_node: Optional[Any]
+    extension_ast_nodes: Tuple[Any, ...]
 
 
 class GraphQLNamedType(GraphQLType):
@@ -194,9 +223,9 @@ class GraphQLNamedType(GraphQLType):
 
     name: str
     description: Optional[str]
-    extensions: Optional[Dict[str, Any]]
+    extensions: Dict[str, Any]
     ast_node: Optional[TypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[TypeExtensionNode]]
+    extension_ast_nodes: Tuple[TypeExtensionNode, ...]
 
     def __init__(
         self,
@@ -206,15 +235,13 @@ class GraphQLNamedType(GraphQLType):
         ast_node: Optional[TypeDefinitionNode] = None,
         extension_ast_nodes: Optional[Collection[TypeExtensionNode]] = None,
     ) -> None:
-        if not name:
-            raise TypeError("Must provide name.")
-        if not isinstance(name, str):
-            raise TypeError("The name must be a string.")
+        assert_name(name)
         if description is not None and not is_description(description):
             raise TypeError("The description must be a string.")
-        if extensions is not None and (
-            not isinstance(extensions, dict)
-            or not all(isinstance(key, str) for key in extensions)
+        if extensions is None:
+            extensions = {}
+        elif not isinstance(extensions, dict) or not all(
+            isinstance(key, str) for key in extensions
         ):
             raise TypeError(f"{name} extensions must be a dictionary with string keys.")
         if ast_node and not isinstance(ast_node, TypeDefinitionNode):
@@ -227,10 +254,10 @@ class GraphQLNamedType(GraphQLType):
                     f"{name} extension AST nodes must be specified"
                     " as a collection of TypeExtensionNode instances."
                 )
-            if not isinstance(extension_ast_nodes, FrozenList):
-                extension_ast_nodes = FrozenList(extension_ast_nodes)
+            if not isinstance(extension_ast_nodes, tuple):
+                extension_ast_nodes = tuple(extension_ast_nodes)
         else:
-            extension_ast_nodes = None
+            extension_ast_nodes = ()
         self.name = name
         self.description = description
         self.extensions = extensions
@@ -243,51 +270,27 @@ class GraphQLNamedType(GraphQLType):
     def __str__(self) -> str:
         return self.name
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
+    def to_kwargs(self) -> GraphQLNamedTypeKwargs:
+        return GraphQLNamedTypeKwargs(
             name=self.name,
             description=self.description,
             extensions=self.extensions,
             ast_node=self.ast_node,
-            extension_ast_nodes=self.extension_ast_nodes or FrozenList(),
+            extension_ast_nodes=self.extension_ast_nodes,
         )
 
     def __copy__(self) -> "GraphQLNamedType":  # pragma: no cover
         return self.__class__(**self.to_kwargs())
 
 
-def is_named_type(type_: Any) -> bool:
-    return isinstance(type_, GraphQLNamedType)
+T = TypeVar("T")
+
+ThunkCollection = Union[Callable[[], Collection[T]], Collection[T]]
+ThunkMapping = Union[Callable[[], Mapping[str, T]], Mapping[str, T]]
+Thunk = Union[Callable[[], T], T]
 
 
-def assert_named_type(type_: Any) -> GraphQLNamedType:
-    if not is_named_type(type_):
-        raise TypeError(f"Expected {type_} to be a GraphQL named type.")
-    return cast(GraphQLNamedType, type_)
-
-
-@overload
-def get_named_type(type_: None) -> None:
-    ...
-
-
-@overload
-def get_named_type(type_: GraphQLType) -> GraphQLNamedType:
-    ...
-
-
-def get_named_type(type_: Optional[GraphQLType]) -> Optional[GraphQLNamedType]:
-    """Unwrap possible wrapping type"""
-    if type_:
-        unwrapped_type = type_
-        while is_wrapping_type(unwrapped_type):
-            unwrapped_type = cast(GraphQLWrappingType, unwrapped_type)
-            unwrapped_type = unwrapped_type.of_type
-        return cast(GraphQLNamedType, unwrapped_type)
-    return None
-
-
-def resolve_thunk(thunk: Any) -> Any:
+def resolve_thunk(thunk: Thunk[T]) -> T:
     """Resolve the given thunk.
 
     Used while defining GraphQL types to allow for circular references in otherwise
@@ -301,6 +304,13 @@ GraphQLScalarValueParser = Callable[[Any], Any]
 GraphQLScalarLiteralParser = Callable[[ValueNode, Optional[Dict[str, Any]]], Any]
 
 
+class GraphQLScalarTypeKwargs(GraphQLNamedTypeKwargs, total=False):
+    serialize: Optional[GraphQLScalarSerializer]
+    parse_value: Optional[GraphQLScalarValueParser]
+    parse_literal: Optional[GraphQLScalarLiteralParser]
+    specified_by_url: Optional[str]
+
+
 class GraphQLScalarType(GraphQLNamedType):
     """Scalar Type Definition
 
@@ -323,7 +333,7 @@ class GraphQLScalarType(GraphQLNamedType):
 
     specified_by_url: Optional[str]
     ast_node: Optional[ScalarTypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[ScalarTypeExtensionNode]]
+    extension_ast_nodes: Tuple[ScalarTypeExtensionNode, ...]
 
     def __init__(
         self,
@@ -415,9 +425,10 @@ class GraphQLScalarType(GraphQLNamedType):
         """
         return self.parse_value(value_from_ast_untyped(node, variables))
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
-            **super().to_kwargs(),
+    def to_kwargs(self) -> GraphQLScalarTypeKwargs:
+        # noinspection PyArgumentList
+        return GraphQLScalarTypeKwargs(  # type: ignore
+            super().to_kwargs(),
             serialize=None
             if self.serialize is GraphQLScalarType.serialize
             else self.serialize,
@@ -448,6 +459,17 @@ def assert_scalar_type(type_: Any) -> GraphQLScalarType:
 GraphQLArgumentMap = Dict[str, "GraphQLArgument"]
 
 
+class GraphQLFieldKwargs(TypedDict, total=False):
+    type_: "GraphQLOutputType"
+    args: Optional[GraphQLArgumentMap]
+    resolve: Optional["GraphQLFieldResolver"]
+    subscribe: Optional["GraphQLFieldResolver"]
+    description: Optional[str]
+    deprecation_reason: Optional[str]
+    extensions: Dict[str, Any]
+    ast_node: Optional[FieldDefinitionNode]
+
+
 class GraphQLField:
     """Definition of a GraphQL field"""
 
@@ -457,7 +479,7 @@ class GraphQLField:
     subscribe: Optional["GraphQLFieldResolver"]
     description: Optional[str]
     deprecation_reason: Optional[str]
-    extensions: Optional[Dict[str, Any]]
+    extensions: Dict[str, Any]
     ast_node: Optional[FieldDefinitionNode]
 
     def __init__(
@@ -486,7 +508,7 @@ class GraphQLField:
             )
         else:
             args = {
-                name: value
+                assert_name(name): value
                 if isinstance(value, GraphQLArgument)
                 else GraphQLArgument(cast(GraphQLInputType, value))
                 for name, value in args.items()
@@ -500,9 +522,10 @@ class GraphQLField:
             raise TypeError("The description must be a string.")
         if deprecation_reason is not None and not is_description(deprecation_reason):
             raise TypeError("The deprecation reason must be a string.")
-        if extensions is not None and (
-            not isinstance(extensions, dict)
-            or not all(isinstance(key, str) for key in extensions)
+        if extensions is None:
+            extensions = {}
+        elif not isinstance(extensions, dict) or not all(
+            isinstance(key, str) for key in extensions
         ):
             raise TypeError("Field extensions must be a dictionary with string keys.")
         if ast_node and not isinstance(ast_node, FieldDefinitionNode):
@@ -533,8 +556,8 @@ class GraphQLField:
             and self.extensions == other.extensions
         )
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
+    def to_kwargs(self) -> GraphQLFieldKwargs:
+        return GraphQLFieldKwargs(
             type_=self.type,
             args=self.args.copy() if self.args else None,
             resolve=self.resolve,
@@ -548,11 +571,6 @@ class GraphQLField:
     def __copy__(self) -> "GraphQLField":  # pragma: no cover
         return self.__class__(**self.to_kwargs())
 
-    @property
-    def is_deprecated(self) -> bool:
-        # this property is officially deprecated, but we still keep it here
-        return self.deprecation_reason is not None
-
 
 class GraphQLResolveInfo(NamedTuple):
     """Collection of information passed to the resolvers.
@@ -565,7 +583,7 @@ class GraphQLResolveInfo(NamedTuple):
     """
 
     field_name: str
-    field_nodes: List[FieldNode]
+    field_nodes: Collection[FieldNode]
     return_type: "GraphQLOutputType"
     parent_type: "GraphQLObjectType"
     path: Path
@@ -588,16 +606,27 @@ GraphQLFieldResolver = Callable[..., Any]
 
 # Note: Contrary to the Javascript implementation of GraphQLTypeResolver,
 # the context is passed as part of the GraphQLResolveInfo:
-# Note: returning GraphQLObjectType is deprecated and will be removed.
 GraphQLTypeResolver = Callable[
     [Any, GraphQLResolveInfo, "GraphQLAbstractType"],
-    AwaitableOrValue[Optional[Union["GraphQLObjectType", str]]],
+    AwaitableOrValue[Optional[str]],
 ]
 
 # Note: Contrary to the Javascript implementation of GraphQLIsTypeOfFn,
 # the context is passed as part of the GraphQLResolveInfo:
 GraphQLIsTypeOfFn = Callable[[Any, GraphQLResolveInfo], AwaitableOrValue[bool]]
 
+GraphQLFieldMap = Dict[str, GraphQLField]
+
+
+class GraphQLArgumentKwargs(TypedDict, total=False):
+    type_: "GraphQLInputType"
+    default_value: Any
+    description: Optional[str]
+    deprecation_reason: Optional[str]
+    out_name: Optional[str]
+    extensions: Dict[str, Any]
+    ast_node: Optional[InputValueDefinitionNode]
+
 
 class GraphQLArgument:
     """Definition of a GraphQL argument"""
@@ -607,7 +636,7 @@ class GraphQLArgument:
     description: Optional[str]
     deprecation_reason: Optional[str]
     out_name: Optional[str]  # for transforming names (extension of GraphQL.js)
-    extensions: Optional[Dict[str, Any]]
+    extensions: Dict[str, Any]
     ast_node: Optional[InputValueDefinitionNode]
 
     def __init__(
@@ -628,9 +657,10 @@ class GraphQLArgument:
             raise TypeError("Argument deprecation reason must be a string.")
         if out_name is not None and not isinstance(out_name, str):
             raise TypeError("Argument out name must be a string.")
-        if extensions is not None and (
-            not isinstance(extensions, dict)
-            or not all(isinstance(key, str) for key in extensions)
+        if extensions is None:
+            extensions = {}
+        elif not isinstance(extensions, dict) or not all(
+            isinstance(key, str) for key in extensions
         ):
             raise TypeError(
                 "Argument extensions must be a dictionary with string keys."
@@ -656,8 +686,8 @@ class GraphQLArgument:
             and self.extensions == other.extensions
         )
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
+    def to_kwargs(self) -> GraphQLArgumentKwargs:
+        return GraphQLArgumentKwargs(
             type_=self.type,
             default_value=self.default_value,
             description=self.description,
@@ -675,16 +705,17 @@ def is_required_argument(arg: GraphQLArgument) -> bool:
     return is_non_null_type(arg.type) and arg.default_value is Undefined
 
 
-T = TypeVar("T")
-Thunk = Union[Callable[[], T], T]
+class GraphQLObjectTypeKwargs(GraphQLNamedTypeKwargs, total=False):
 
-GraphQLFieldMap = Dict[str, GraphQLField]
+    fields: GraphQLFieldMap
+    interfaces: Tuple["GraphQLInterfaceType", ...]
+    is_type_of: Optional[GraphQLIsTypeOfFn]
 
 
 class GraphQLObjectType(GraphQLNamedType):
     """Object Type Definition
 
-    Almost all of the GraphQL types you define will be object types. Object types have
+    Almost all the GraphQL types you define will be object types. Object types have
     a name, but most importantly describe their fields.
 
     Example::
@@ -711,13 +742,13 @@ class GraphQLObjectType(GraphQLNamedType):
 
     is_type_of: Optional[GraphQLIsTypeOfFn]
     ast_node: Optional[ObjectTypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[ObjectTypeExtensionNode]]
+    extension_ast_nodes: Tuple[ObjectTypeExtensionNode, ...]
 
     def __init__(
         self,
         name: str,
-        fields: Thunk[GraphQLFieldMap],
-        interfaces: Optional[Thunk[Collection["GraphQLInterfaceType"]]] = None,
+        fields: ThunkMapping[GraphQLField],
+        interfaces: Optional[ThunkCollection["GraphQLInterfaceType"]] = None,
         is_type_of: Optional[GraphQLIsTypeOfFn] = None,
         extensions: Optional[Dict[str, Any]] = None,
         description: Optional[str] = None,
@@ -749,9 +780,10 @@ class GraphQLObjectType(GraphQLNamedType):
         self._interfaces = interfaces
         self.is_type_of = is_type_of
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
-            **super().to_kwargs(),
+    def to_kwargs(self) -> GraphQLObjectTypeKwargs:
+        # noinspection PyArgumentList
+        return GraphQLObjectTypeKwargs(  # type: ignore
+            super().to_kwargs(),
             fields=self.fields.copy(),
             interfaces=self.interfaces,
             is_type_of=self.is_type_of,
@@ -766,13 +798,14 @@ class GraphQLObjectType(GraphQLNamedType):
         try:
             fields = resolve_thunk(self._fields)
         except Exception as error:
-            raise TypeError(f"{self.name} fields cannot be resolved. {error}")
-        if not isinstance(fields, dict) or not all(
+            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
+            raise cls(f"{self.name} fields cannot be resolved. {error}") from error
+        if not isinstance(fields, Mapping) or not all(
             isinstance(key, str) for key in fields
         ):
             raise TypeError(
                 f"{self.name} fields must be specified"
-                " as a dict with field names as keys."
+                " as a mapping with field names as keys."
             )
         if not all(
             isinstance(value, GraphQLField) or is_output_type(value)
@@ -782,21 +815,24 @@ class GraphQLObjectType(GraphQLNamedType):
                 f"{self.name} fields must be GraphQLField or output type objects."
             )
         return {
-            name: value if isinstance(value, GraphQLField) else GraphQLField(value)
+            assert_name(name): value
+            if isinstance(value, GraphQLField)
+            else GraphQLField(value)  # type: ignore
             for name, value in fields.items()
         }
 
     @cached_property
-    def interfaces(self) -> List["GraphQLInterfaceType"]:
+    def interfaces(self) -> Tuple["GraphQLInterfaceType", ...]:
         """Get provided interfaces."""
         try:
             interfaces: Collection["GraphQLInterfaceType"] = resolve_thunk(
-                self._interfaces
+                self._interfaces  # type: ignore
             )
         except Exception as error:
-            raise TypeError(f"{self.name} interfaces cannot be resolved. {error}")
+            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
+            raise cls(f"{self.name} interfaces cannot be resolved. {error}") from error
         if interfaces is None:
-            interfaces = []
+            interfaces = ()
         elif not is_collection(interfaces) or not all(
             isinstance(value, GraphQLInterfaceType) for value in interfaces
         ):
@@ -804,7 +840,7 @@ class GraphQLObjectType(GraphQLNamedType):
                 f"{self.name} interfaces must be specified"
                 " as a collection of GraphQLInterfaceType instances."
             )
-        return list(interfaces)
+        return tuple(interfaces)
 
 
 def is_object_type(type_: Any) -> bool:
@@ -817,6 +853,12 @@ def assert_object_type(type_: Any) -> GraphQLObjectType:
     return cast(GraphQLObjectType, type_)
 
 
+class GraphQLInterfaceTypeKwargs(GraphQLNamedTypeKwargs, total=False):
+    fields: GraphQLFieldMap
+    interfaces: Tuple["GraphQLInterfaceType", ...]
+    resolve_type: Optional[GraphQLTypeResolver]
+
+
 class GraphQLInterfaceType(GraphQLNamedType):
     """Interface Type Definition
 
@@ -834,13 +876,13 @@ class GraphQLInterfaceType(GraphQLNamedType):
 
     resolve_type: Optional[GraphQLTypeResolver]
     ast_node: Optional[InterfaceTypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[InterfaceTypeExtensionNode]]
+    extension_ast_nodes: Tuple[InterfaceTypeExtensionNode, ...]
 
     def __init__(
         self,
         name: str,
-        fields: Optional[Thunk[GraphQLFieldMap]] = None,
-        interfaces: Optional[Thunk[Collection["GraphQLInterfaceType"]]] = None,
+        fields: ThunkMapping[GraphQLField],
+        interfaces: Optional[ThunkCollection["GraphQLInterfaceType"]] = None,
         resolve_type: Optional[GraphQLTypeResolver] = None,
         description: Optional[str] = None,
         extensions: Optional[Dict[str, Any]] = None,
@@ -872,11 +914,12 @@ class GraphQLInterfaceType(GraphQLNamedType):
         self._interfaces = interfaces
         self.resolve_type = resolve_type
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
-            **super().to_kwargs(),
+    def to_kwargs(self) -> GraphQLInterfaceTypeKwargs:
+        # noinspection PyArgumentList
+        return GraphQLInterfaceTypeKwargs(  # type: ignore
+            super().to_kwargs(),
             fields=self.fields.copy(),
-            interfaces=self.interfaces.copy(),
+            interfaces=self.interfaces,
             resolve_type=self.resolve_type,
         )
 
@@ -889,13 +932,14 @@ class GraphQLInterfaceType(GraphQLNamedType):
         try:
             fields = resolve_thunk(self._fields)
         except Exception as error:
-            raise TypeError(f"{self.name} fields cannot be resolved. {error}")
-        if not isinstance(fields, dict) or not all(
+            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
+            raise cls(f"{self.name} fields cannot be resolved. {error}") from error
+        if not isinstance(fields, Mapping) or not all(
             isinstance(key, str) for key in fields
         ):
             raise TypeError(
                 f"{self.name} fields must be specified"
-                " as a dict with field names as keys."
+                " as a mapping with field names as keys."
             )
         if not all(
             isinstance(value, GraphQLField) or is_output_type(value)
@@ -905,21 +949,24 @@ class GraphQLInterfaceType(GraphQLNamedType):
                 f"{self.name} fields must be GraphQLField or output type objects."
             )
         return {
-            name: value if isinstance(value, GraphQLField) else GraphQLField(value)
+            assert_name(name): value
+            if isinstance(value, GraphQLField)
+            else GraphQLField(value)  # type: ignore
             for name, value in fields.items()
         }
 
     @cached_property
-    def interfaces(self) -> List["GraphQLInterfaceType"]:
+    def interfaces(self) -> Tuple["GraphQLInterfaceType", ...]:
         """Get provided interfaces."""
         try:
             interfaces: Collection["GraphQLInterfaceType"] = resolve_thunk(
-                self._interfaces
+                self._interfaces  # type: ignore
             )
         except Exception as error:
-            raise TypeError(f"{self.name} interfaces cannot be resolved. {error}")
+            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
+            raise cls(f"{self.name} interfaces cannot be resolved. {error}") from error
         if interfaces is None:
-            interfaces = []
+            interfaces = ()
         elif not is_collection(interfaces) or not all(
             isinstance(value, GraphQLInterfaceType) for value in interfaces
         ):
@@ -927,7 +974,7 @@ class GraphQLInterfaceType(GraphQLNamedType):
                 f"{self.name} interfaces must be specified"
                 " as a collection of GraphQLInterfaceType instances."
             )
-        return list(interfaces)
+        return tuple(interfaces)
 
 
 def is_interface_type(type_: Any) -> bool:
@@ -940,6 +987,11 @@ def assert_interface_type(type_: Any) -> GraphQLInterfaceType:
     return cast(GraphQLInterfaceType, type_)
 
 
+class GraphQLUnionTypeKwargs(GraphQLNamedTypeKwargs, total=False):
+    types: Tuple[GraphQLObjectType, ...]
+    resolve_type: Optional[GraphQLTypeResolver]
+
+
 class GraphQLUnionType(GraphQLNamedType):
     """Union Type Definition
 
@@ -960,12 +1012,12 @@ class GraphQLUnionType(GraphQLNamedType):
 
     resolve_type: Optional[GraphQLTypeResolver]
     ast_node: Optional[UnionTypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[UnionTypeExtensionNode]]
+    extension_ast_nodes: Tuple[UnionTypeExtensionNode, ...]
 
     def __init__(
         self,
         name: str,
-        types: Thunk[Collection[GraphQLObjectType]],
+        types: ThunkCollection[GraphQLObjectType],
         resolve_type: Optional[GraphQLTypeResolver] = None,
         description: Optional[str] = None,
         extensions: Optional[Dict[str, Any]] = None,
@@ -996,23 +1048,25 @@ class GraphQLUnionType(GraphQLNamedType):
         self._types = types
         self.resolve_type = resolve_type
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
-            **super().to_kwargs(), types=self.types, resolve_type=self.resolve_type
+    def to_kwargs(self) -> GraphQLUnionTypeKwargs:
+        # noinspection PyArgumentList
+        return GraphQLUnionTypeKwargs(  # type: ignore
+            super().to_kwargs(), types=self.types, resolve_type=self.resolve_type
         )
 
     def __copy__(self) -> "GraphQLUnionType":  # pragma: no cover
         return self.__class__(**self.to_kwargs())
 
     @cached_property
-    def types(self) -> List[GraphQLObjectType]:
+    def types(self) -> Tuple[GraphQLObjectType, ...]:
         """Get provided types."""
         try:
             types: Collection[GraphQLObjectType] = resolve_thunk(self._types)
         except Exception as error:
-            raise TypeError(f"{self.name} types cannot be resolved. {error}")
+            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
+            raise cls(f"{self.name} types cannot be resolved. {error}") from error
         if types is None:
-            types = []
+            types = ()
         elif not is_collection(types) or not all(
             isinstance(value, GraphQLObjectType) for value in types
         ):
@@ -1020,7 +1074,7 @@ class GraphQLUnionType(GraphQLNamedType):
                 f"{self.name} types must be specified"
                 " as a collection of GraphQLObjectType instances."
             )
-        return list(types)
+        return tuple(types)
 
 
 def is_union_type(type_: Any) -> bool:
@@ -1036,12 +1090,20 @@ def assert_union_type(type_: Any) -> GraphQLUnionType:
 GraphQLEnumValueMap = Dict[str, "GraphQLEnumValue"]
 
 
+class GraphQLEnumTypeKwargs(GraphQLNamedTypeKwargs, total=False):
+    values: GraphQLEnumValueMap
+    names_as_values: Optional[bool]
+
+
 class GraphQLEnumType(GraphQLNamedType):
     """Enum Type Definition
 
     Some leaf values of requests and input values are Enums. GraphQL serializes Enum
     values as strings, however internally Enums can be represented by any kind of type,
-    often integers. They can also be provided as a Python Enum.
+    often integers. They can also be provided as a Python Enum. In this case, the flag
+    `names_as_values` determines what will be used as internal representation. The
+    default value of `False` will use the enum values, the value `True` will use the
+    enum names, and the value `None` will use the members themselves.
 
     Example::
 
@@ -1069,12 +1131,13 @@ class GraphQLEnumType(GraphQLNamedType):
 
     values: GraphQLEnumValueMap
     ast_node: Optional[EnumTypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[EnumTypeExtensionNode]]
+    extension_ast_nodes: Tuple[EnumTypeExtensionNode, ...]
 
     def __init__(
         self,
         name: str,
-        values: Union[GraphQLEnumValueMap, Dict[str, Any], Type[Enum]],
+        values: Union[GraphQLEnumValueMap, Mapping[str, Any], Type[Enum]],
+        names_as_values: Optional[bool] = False,
         description: Optional[str] = None,
         extensions: Optional[Dict[str, Any]] = None,
         ast_node: Optional[EnumTypeDefinitionNode] = None,
@@ -1090,7 +1153,7 @@ class GraphQLEnumType(GraphQLNamedType):
         try:  # check for enum
             values = cast(Enum, values).__members__  # type: ignore
         except AttributeError:
-            if not isinstance(values, dict) or not all(
+            if not isinstance(values, Mapping) or not all(
                 isinstance(name, str) for name in values
             ):
                 try:
@@ -1098,15 +1161,18 @@ class GraphQLEnumType(GraphQLNamedType):
                     values = dict(values)  # type: ignore
                 except (TypeError, ValueError):
                     raise TypeError(
-                        f"{name} values must be an Enum or a dict"
+                        f"{name} values must be an Enum or a mapping"
                         " with value names as keys."
                     )
-            values = cast(Dict, values)
+            values = cast(Dict[str, Any], values)
         else:
-            values = cast(Dict, values)
-            values = {key: value.value for key, value in values.items()}
+            values = cast(Dict[str, Enum], values)
+            if names_as_values is False:
+                values = {key: value.value for key, value in values.items()}
+            elif names_as_values is True:
+                values = {key: key for key in values}
         values = {
-            key: value
+            assert_enum_value_name(key): value
             if isinstance(value, GraphQLEnumValue)
             else GraphQLEnumValue(value)
             for key, value in values.items()
@@ -1122,8 +1188,11 @@ class GraphQLEnumType(GraphQLNamedType):
             )
         self.values = values
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(**super().to_kwargs(), values=self.values.copy())
+    def to_kwargs(self) -> GraphQLEnumTypeKwargs:
+        # noinspection PyArgumentList
+        return GraphQLEnumTypeKwargs(  # type: ignore
+            super().to_kwargs(), values=self.values.copy()
+        )
 
     def __copy__(self) -> "GraphQLEnumType":  # pragma: no cover
         return self.__class__(**self.to_kwargs())
@@ -1210,12 +1279,20 @@ def did_you_mean_enum_value(enum_type: GraphQLEnumType, unknown_value_str: str)
     return did_you_mean(suggested_values, "the enum value")
 
 
+class GraphQLEnumValueKwargs(TypedDict, total=False):
+    value: Any
+    description: Optional[str]
+    deprecation_reason: Optional[str]
+    extensions: Dict[str, Any]
+    ast_node: Optional[EnumValueDefinitionNode]
+
+
 class GraphQLEnumValue:
 
     value: Any
     description: Optional[str]
     deprecation_reason: Optional[str]
-    extensions: Optional[Dict[str, Any]]
+    extensions: Dict[str, Any]
     ast_node: Optional[EnumValueDefinitionNode]
 
     def __init__(
@@ -1232,9 +1309,10 @@ class GraphQLEnumValue:
             raise TypeError(
                 "The deprecation reason for the enum value must be a string."
             )
-        if extensions is not None and (
-            not isinstance(extensions, dict)
-            or not all(isinstance(key, str) for key in extensions)
+        if extensions is None:
+            extensions = {}
+        elif not isinstance(extensions, dict) or not all(
+            isinstance(key, str) for key in extensions
         ):
             raise TypeError(
                 "Enum value extensions must be a dictionary with string keys."
@@ -1256,8 +1334,8 @@ class GraphQLEnumValue:
             and self.extensions == other.extensions
         )
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
+    def to_kwargs(self) -> GraphQLEnumValueKwargs:
+        return GraphQLEnumValueKwargs(
             value=self.value,
             description=self.description,
             deprecation_reason=self.deprecation_reason,
@@ -1268,16 +1346,16 @@ class GraphQLEnumValue:
     def __copy__(self) -> "GraphQLEnumValue":  # pragma: no cover
         return self.__class__(**self.to_kwargs())
 
-    @property
-    def is_deprecated(self) -> bool:
-        # this property is officially deprecated, but we still keep it here
-        return self.deprecation_reason is not None
-
 
 GraphQLInputFieldMap = Dict[str, "GraphQLInputField"]
 GraphQLInputFieldOutType = Callable[[Dict[str, Any]], Any]
 
 
+class GraphQLInputObjectTypeKwargs(GraphQLNamedTypeKwargs, total=False):
+    fields: GraphQLInputFieldMap
+    out_type: Optional[GraphQLInputFieldOutType]
+
+
 class GraphQLInputObjectType(GraphQLNamedType):
     """Input Object Type Definition
 
@@ -1304,12 +1382,12 @@ class GraphQLInputObjectType(GraphQLNamedType):
     """
 
     ast_node: Optional[InputObjectTypeDefinitionNode]
-    extension_ast_nodes: Optional[FrozenList[InputObjectTypeExtensionNode]]
+    extension_ast_nodes: Tuple[InputObjectTypeExtensionNode, ...]
 
     def __init__(
         self,
         name: str,
-        fields: Thunk[GraphQLInputFieldMap],
+        fields: ThunkMapping["GraphQLInputField"],
         description: Optional[str] = None,
         out_type: Optional[GraphQLInputFieldOutType] = None,
         extensions: Optional[Dict[str, Any]] = None,
@@ -1349,9 +1427,10 @@ class GraphQLInputObjectType(GraphQLNamedType):
         """
         return value
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
-            **super().to_kwargs(),
+    def to_kwargs(self) -> GraphQLInputObjectTypeKwargs:
+        # noinspection PyArgumentList
+        return GraphQLInputObjectTypeKwargs(  # type: ignore
+            super().to_kwargs(),
             fields=self.fields.copy(),
             out_type=None
             if self.out_type is GraphQLInputObjectType.out_type
@@ -1367,13 +1446,14 @@ class GraphQLInputObjectType(GraphQLNamedType):
         try:
             fields = resolve_thunk(self._fields)
         except Exception as error:
-            raise TypeError(f"{self.name} fields cannot be resolved. {error}")
-        if not isinstance(fields, dict) or not all(
+            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
+            raise cls(f"{self.name} fields cannot be resolved. {error}") from error
+        if not isinstance(fields, Mapping) or not all(
             isinstance(key, str) for key in fields
         ):
             raise TypeError(
                 f"{self.name} fields must be specified"
-                " as a dict with field names as keys."
+                " as a mapping with field names as keys."
             )
         if not all(
             isinstance(value, GraphQLInputField) or is_input_type(value)
@@ -1384,9 +1464,9 @@ class GraphQLInputObjectType(GraphQLNamedType):
                 " GraphQLInputField or input type objects."
             )
         return {
-            name: value
+            assert_name(name): value
             if isinstance(value, GraphQLInputField)
-            else GraphQLInputField(value)
+            else GraphQLInputField(value)  # type: ignore
             for name, value in fields.items()
         }
 
@@ -1401,6 +1481,16 @@ def assert_input_object_type(type_: Any) -> GraphQLInputObjectType:
     return cast(GraphQLInputObjectType, type_)
 
 
+class GraphQLInputFieldKwargs(TypedDict, total=False):
+    type_: "GraphQLInputType"
+    default_value: Any
+    description: Optional[str]
+    deprecation_reason: Optional[str]
+    out_name: Optional[str]
+    extensions: Dict[str, Any]
+    ast_node: Optional[InputValueDefinitionNode]
+
+
 class GraphQLInputField:
     """Definition of a GraphQL input field"""
 
@@ -1409,7 +1499,7 @@ class GraphQLInputField:
     description: Optional[str]
     deprecation_reason: Optional[str]
     out_name: Optional[str]  # for transforming names (extension of GraphQL.js)
-    extensions: Optional[Dict[str, Any]]
+    extensions: Dict[str, Any]
     ast_node: Optional[InputValueDefinitionNode]
 
     def __init__(
@@ -1430,9 +1520,10 @@ class GraphQLInputField:
             raise TypeError("Input field deprecation reason must be a string.")
         if out_name is not None and not isinstance(out_name, str):
             raise TypeError("Input field out name must be a string.")
-        if extensions is not None and (
-            not isinstance(extensions, dict)
-            or not all(isinstance(key, str) for key in extensions)
+        if extensions is None:
+            extensions = {}
+        elif not isinstance(extensions, dict) or not all(
+            isinstance(key, str) for key in extensions
         ):
             raise TypeError(
                 "Input field extensions must be a dictionary with string keys."
@@ -1458,8 +1549,8 @@ class GraphQLInputField:
             and self.out_name == other.out_name
         )
 
-    def to_kwargs(self) -> Dict[str, Any]:
-        return dict(
+    def to_kwargs(self) -> GraphQLInputFieldKwargs:
+        return GraphQLInputFieldKwargs(
             type_=self.type,
             default_value=self.default_value,
             description=self.description,
@@ -1672,6 +1763,52 @@ def assert_output_type(type_: Any) -> GraphQLOutputType:
     return cast(GraphQLOutputType, type_)
 
 
+# These named types do not include modifiers like List or NonNull.
+
+GraphQLNamedInputType = Union[
+    GraphQLScalarType, GraphQLEnumType, GraphQLInputObjectType
+]
+
+GraphQLNamedOutputType = Union[
+    GraphQLScalarType,
+    GraphQLObjectType,
+    GraphQLInterfaceType,
+    GraphQLUnionType,
+    GraphQLEnumType,
+]
+
+
+def is_named_type(type_: Any) -> bool:
+    return isinstance(type_, GraphQLNamedType)
+
+
+def assert_named_type(type_: Any) -> GraphQLNamedType:
+    if not is_named_type(type_):
+        raise TypeError(f"Expected {type_} to be a GraphQL named type.")
+    return cast(GraphQLNamedType, type_)
+
+
+@overload
+def get_named_type(type_: None) -> None:
+    ...
+
+
+@overload
+def get_named_type(type_: GraphQLType) -> GraphQLNamedType:
+    ...
+
+
+def get_named_type(type_: Optional[GraphQLType]) -> Optional[GraphQLNamedType]:
+    """Unwrap possible wrapping type"""
+    if type_:
+        unwrapped_type = type_
+        while is_wrapping_type(unwrapped_type):
+            unwrapped_type = cast(GraphQLWrappingType, unwrapped_type)
+            unwrapped_type = unwrapped_type.of_type
+        return cast(GraphQLNamedType, unwrapped_type)
+    return None
+
+
 # These types may describe types which may be leaf values.
 
 graphql_leaf_types = (GraphQLScalarType, GraphQLEnumType)
