@@ -49,6 +49,8 @@ class OverlappingFieldsCanBeMergedRule(ValidationRule):
 
     A selection set is only valid if all fields (including spreading any fragments)
     either correspond to distinct response names or can be merged without ambiguity.
+
+    See https://spec.graphql.org/draft/#sec-Field-Selection-Merging
     """
 
     def __init__(self, context: ValidationContext):
@@ -229,7 +231,7 @@ def collect_conflicts_between_fields_and_fragment(
     if not fragment:
         return None
 
-    field_map2, fragment_names2 = get_referenced_fields_and_fragment_names(
+    field_map2, referenced_fragment_names = get_referenced_fields_and_fragment_names(
         context, cached_fields_and_fragment_names, fragment
     )
 
@@ -251,7 +253,16 @@ def collect_conflicts_between_fields_and_fragment(
 
     # (E) Then collect any conflicts between the provided collection of fields and any
     # fragment names found in the given fragment.
-    for fragment_name2 in fragment_names2:
+    for referenced_fragment_name in referenced_fragment_names:
+        # Memoize so two fragments are not compared for conflicts more than once.
+        if compared_fragment_pairs.has(
+            referenced_fragment_name, fragment_name, are_mutually_exclusive
+        ):
+            continue
+        compared_fragment_pairs.add(
+            referenced_fragment_name, fragment_name, are_mutually_exclusive
+        )
+
         collect_conflicts_between_fields_and_fragment(
             context,
             conflicts,
@@ -259,7 +270,7 @@ def collect_conflicts_between_fields_and_fragment(
             compared_fragment_pairs,
             are_mutually_exclusive,
             field_map,
-            fragment_name2,
+            referenced_fragment_name,
         )
 
 
@@ -293,11 +304,11 @@ def collect_conflicts_between_fragments(
     if not fragment1 or not fragment2:
         return None
 
-    field_map1, fragment_names1 = get_referenced_fields_and_fragment_names(
+    field_map1, referenced_fragment_names1 = get_referenced_fields_and_fragment_names(
         context, cached_fields_and_fragment_names, fragment1
     )
 
-    field_map2, fragment_names2 = get_referenced_fields_and_fragment_names(
+    field_map2, referenced_fragment_names2 = get_referenced_fields_and_fragment_names(
         context, cached_fields_and_fragment_names, fragment2
     )
 
@@ -315,7 +326,7 @@ def collect_conflicts_between_fragments(
 
     # (G) Then collect conflicts between the first fragment and any nested fragments
     # spread in the second fragment.
-    for nested_fragment_name2 in fragment_names2:
+    for referenced_fragment_name2 in referenced_fragment_names2:
         collect_conflicts_between_fragments(
             context,
             conflicts,
@@ -323,19 +334,19 @@ def collect_conflicts_between_fragments(
             compared_fragment_pairs,
             are_mutually_exclusive,
             fragment_name1,
-            nested_fragment_name2,
+            referenced_fragment_name2,
         )
 
     # (G) Then collect conflicts between the second fragment and any nested fragments
     # spread in the first fragment.
-    for nested_fragment_name1 in fragment_names1:
+    for referenced_fragment_name1 in referenced_fragment_names1:
         collect_conflicts_between_fragments(
             context,
             conflicts,
             cached_fields_and_fragment_names,
             compared_fragment_pairs,
             are_mutually_exclusive,
-            nested_fragment_name1,
+            referenced_fragment_name1,
             fragment_name2,
         )
 
@@ -726,34 +737,36 @@ def subfield_conflicts(
 class PairSet:
     """Pair set
 
-    A way to keep track of pairs of things when the ordering of the pair does not
-    matter. We do this by maintaining a sort of double adjacency sets.
+    A way to keep track of pairs of things when the ordering of the pair doesn't matter.
     """
 
     __slots__ = ("_data",)
 
+    _data: Dict[str, Dict[str, bool]]
+
     def __init__(self) -> None:
-        self._data: Dict[str, Dict[str, bool]] = {}
+        self._data = {}
 
     def has(self, a: str, b: str, are_mutually_exclusive: bool) -> bool:
-        first = self._data.get(a)
-        result = first and first.get(b)
+        key1, key2 = (a, b) if a < b else (b, a)
+
+        map_ = self._data.get(key1)
+        if map_ is None:
+            return False
+        result = map_.get(key2)
         if result is None:
             return False
-        # `are_mutually_exclusive` being False is a superset of being True, hence if we
-        # want to know if this PairSet "has" these two with no exclusivity, we have to
-        # ensure it was added as such.
-        if not are_mutually_exclusive:
-            return not result
-        return True
 
-    def add(self, a: str, b: str, are_mutually_exclusive: bool) -> "PairSet":
-        self._pair_set_add(a, b, are_mutually_exclusive)
-        self._pair_set_add(b, a, are_mutually_exclusive)
-        return self
+        # are_mutually_exclusive being False is a superset of being True,
+        # hence if we want to know if this PairSet "has" these two with no exclusivity,
+        # we have to ensure it was added as such.
+        return True if are_mutually_exclusive else are_mutually_exclusive == result
 
-    def _pair_set_add(self, a: str, b: str, are_mutually_exclusive: bool) -> None:
-        a_map = self._data.get(a)
-        if not a_map:
-            self._data[a] = a_map = {}
-        a_map[b] = are_mutually_exclusive
+    def add(self, a: str, b: str, are_mutually_exclusive: bool) -> None:
+        key1, key2 = (a, b) if a < b else (b, a)
+
+        map_ = self._data.get(key1)
+        if map_ is None:
+            self._data[key1] = {key2: are_mutually_exclusive}
+        else:
+            map_[key2] = are_mutually_exclusive
